
/***
  Copyright (C) 2014 Elementary Calculator Lars Berscheid
  This program is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License version 3, as published
  by the Free Software Foundation.
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranties of
  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along
  with this program. If not, write to the
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
***/


namespace Calculator {

    // possible: public class MainWindow : Granite.Widgets.LightWindow {
    public class MainWindow : Gtk.Window {

        private struct history_entry { string form; string resu; }


        private CalculatorApp app;
        private Parser parser;

        private Gtk.TextView textview;
        private Gtk.Menu menu;
        private Gtk.Button button_history;

        private string formula = "";
        private bool input_new_formula;
        private bool formula_has_error;
        private Array<history_entry?> history;

        // Decimal Symbol can also be ","
        private string decimal_symbol = ".";


        public MainWindow (CalculatorApp app) {
            this.app = app;
            this.set_application (app);

            this.set_size_request (400, 100);
            this.window_position = Gtk.WindowPosition.CENTER;
            this.set_resizable (false);

            parser = new Parser (decimal_symbol);
            menu = new Gtk.Menu ();
            history = new Array<history_entry?> ();

            // Gtk.Settings.get_default().set("gtk-application-prefer-dark-theme", true);

            // Build UI
            build_ui ();
            this.show_all ();

            Gdk.Display display = this.get_display ();
            Gtk.Clipboard clipboard = Gtk.Clipboard.get_for_display (display, Gdk.SELECTION_CLIPBOARD);

            equation_changed ();
            input_new_formula = true;
            set_formula_has_error (false);

            // Key Events
            this.key_press_event.connect ((e) => {
                switch (e.keyval) {
                    case Gdk.Key.@0: append_to_formula ("0"); break;
                    case Gdk.Key.@1: append_to_formula ("1"); break;
                    case Gdk.Key.@2: append_to_formula ("2"); break;
                    case Gdk.Key.@3: append_to_formula ("3"); break;
                    case Gdk.Key.@4: append_to_formula ("4"); break;
                    case Gdk.Key.@5: append_to_formula ("5"); break;
                    case Gdk.Key.@6: append_to_formula ("6"); break;
                    case Gdk.Key.@7: append_to_formula ("7"); break;
                    case Gdk.Key.@8: append_to_formula ("8"); break;
                    case Gdk.Key.@9: append_to_formula ("9"); break;

                    case Gdk.Key.a: append_to_formula ("a"); break;
                    case Gdk.Key.c:
                        if ((e.state & Gdk.ModifierType.CONTROL_MASK) != 0)
                            clipboard.set_text (formula, -1);
                        else
                            append_to_formula ("c");
                        break;
                    case Gdk.Key.d: append_to_formula ("d"); break;
                    case Gdk.Key.e: append_to_formula ("e"); break;
                    case Gdk.Key.g: append_to_formula ("g"); break;
                    case Gdk.Key.h: append_to_formula ("h"); break;
                    case Gdk.Key.i: append_to_formula ("i"); break;
                    case Gdk.Key.l: append_to_formula ("l"); break;
                    case Gdk.Key.m: append_to_formula ("m"); break;
                    case Gdk.Key.n: append_to_formula ("n"); break;
                    case Gdk.Key.o: append_to_formula ("o"); break;
                    case Gdk.Key.p: append_to_formula ("p"); break;
                    case Gdk.Key.q: append_to_formula ("q"); break;
                    case Gdk.Key.r: append_to_formula ("r"); break;
                    case Gdk.Key.s: append_to_formula ("s"); break;
                    case Gdk.Key.t: append_to_formula ("t"); break;
                    case Gdk.Key.v:
                        if ((e.state & Gdk.ModifierType.CONTROL_MASK) != 0)
                            return paste_token (clipboard.wait_for_text ());
                        else
                            return false;
                    case Gdk.Key.x: append_to_formula ("x"); break;
                    case Gdk.Key.z:
                        if ((e.state & Gdk.ModifierType.CONTROL_MASK) != 0)
                            return undo ();
                        else
                            return false;

                    case Gdk.Key.plus: append_to_formula ("+"); break;
                    case Gdk.Key.minus: append_to_formula ("-"); break;
                    case Gdk.Key.asterisk: append_to_formula ("*"); break;
                    case Gdk.Key.slash: append_to_formula ("/"); break;
                    case Gdk.Key.percent: append_to_formula ("%"); break;
                    case Gdk.Key.parenleft: append_to_formula ("("); break;
                    case Gdk.Key.parenright: append_to_formula (")"); break;
                    case Gdk.Key.period:
                        if (decimal_symbol == ".")
                            append_to_formula (".");
                        else
                            return false;
                        break;
                    case Gdk.Key.comma:
                        if (decimal_symbol == ",")
                            append_to_formula (".");
                        else
                            return false;
                        break;

                    case Gdk.Key.Return: solve_equation (); break;
                    case Gdk.Key.equal: solve_equation (); break;
                    case Gdk.Key.Escape: set_formula (""); break;
                    case Gdk.Key.BackSpace:
                        if (formula.length > 0)
                            set_formula (formula[0:-1]);
                        break;

                    default: return false;
                }

                return true;
            });
        }

        private void build_ui () {

            // Header Bar
		    var header = new Gtk.HeaderBar ();
            header.set_show_close_button (true);
            header.set_title ("Calculator");
            header.get_style_context ().remove_class ("header-bar");
            this.set_titlebar (header);

            // Grid, Textview and Buttons
            int left_right_padding = 8, top_buttom_padding = 8;

            var grid = new Gtk.Grid ();
            grid.set_row_homogeneous (true);
            grid.set_column_homogeneous (true);
            grid.set_row_spacing (4);
            grid.set_column_spacing (4);

            var box_h = new Gtk.Box (Gtk.Orientation.HORIZONTAL, 0);
            var box_v = new Gtk.Box (Gtk.Orientation.VERTICAL, 0);
            box_h.pack_start(grid, true, true, left_right_padding);
            box_v.pack_start(box_h, true, true, top_buttom_padding);
            this.add (box_v);

            textview = new Gtk.TextView ();
            textview.set_wrap_mode (Gtk.WrapMode.NONE);
            textview.set_accepts_tab (false);
            textview.set_editable (false);
            textview.set_cursor_visible (false);
            textview.set_justification (Gtk.Justification.RIGHT);
            textview.set_border_window_size (Gtk.TextWindowType.LEFT, 2);
            textview.set_border_window_size (Gtk.TextWindowType.RIGHT, 2);
            textview.set_border_window_size (Gtk.TextWindowType.TOP, 2);
            textview.set_border_window_size (Gtk.TextWindowType.BOTTOM, 2);
            textview.get_style_context ().add_class ("h2");

            var scrolled = new Gtk.ScrolledWindow (null, null);
            scrolled.vscrollbar_policy = Gtk.PolicyType.NEVER;
            scrolled.add (textview);
            scrolled.expand = true;

            var frame = new Gtk.Frame (null);
            frame.add (scrolled);
            grid.attach (frame, 0, 0, 4, 1);

            var button_0 = new Gtk.Button.with_label ("0");
            var button_1 = new Gtk.Button.with_label ("1");
            var button_2 = new Gtk.Button.with_label ("2");
            var button_3 = new Gtk.Button.with_label ("3");
            var button_4 = new Gtk.Button.with_label ("4");
            var button_5 = new Gtk.Button.with_label ("5");
            var button_6 = new Gtk.Button.with_label ("6");
            var button_7 = new Gtk.Button.with_label ("7");
            var button_8 = new Gtk.Button.with_label ("8");
            var button_9 = new Gtk.Button.with_label ("9");

            var button_point = new Gtk.Button.with_label (".");
            var button_equal = new Gtk.Button.with_label ("=");
            var button_plus = new Gtk.Button.with_label ("+");
            var button_minus = new Gtk.Button.with_label ("-");
            var button_multi = new Gtk.Button.with_label ("×");
            var button_div = new Gtk.Button.with_label ("÷");
            var button_percent = new Gtk.Button.with_label ("%");

            button_history = new Gtk.Button (); // .with_label ("H");
            button_history.set_image (new Gtk.Image.from_icon_name ("document-open-recent-symbolic", Gtk.IconSize.MENU));
            button_history.sensitive = false;
            var button_clear = new Gtk.Button (); // .with_label ("C");
            button_clear.set_image (new Gtk.Image.from_icon_name ("edit-clear-symbolic", Gtk.IconSize.MENU));

            button_equal.get_style_context ().add_class (Gtk.STYLE_CLASS_SUGGESTED_ACTION);
            button_clear.get_style_context ().add_class (Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION);
            /* var mybgcolor=Gdk.RGBA();
            mybgcolor.parse("rgba(205,205,205,1)");
            button_clear.override_color(Gtk.StateFlags.NORMAL, mybgcolor); */
            // button_back.get_style_context ().add_class (Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION);

            grid.attach (button_0, 0, 5, 1, 1);
            grid.attach (button_1, 0, 4, 1, 1);
            grid.attach (button_2, 1, 4, 1, 1);
            grid.attach (button_3, 2, 4, 1, 1);
            grid.attach (button_4, 0, 3, 1, 1);
            grid.attach (button_5, 1, 3, 1, 1);
            grid.attach (button_6, 2, 3, 1, 1);
            grid.attach (button_7, 0, 2, 1, 1);
            grid.attach (button_8, 1, 2, 1, 1);
            grid.attach (button_9, 2, 2, 1, 1);

            grid.attach (button_clear, 0, 1, 1, 1);
            grid.attach (button_history, 1, 1, 1, 1);
            grid.attach (button_percent, 2, 1, 1, 1);
            grid.attach (button_point, 1, 5, 1, 1);
            grid.attach (button_equal, 2, 5, 2, 1);
            grid.attach (button_plus, 3, 4, 1, 1);
            grid.attach (button_minus, 3, 3, 1, 1);
            grid.attach (button_multi, 3, 2, 1, 1);
            grid.attach (button_div, 3, 1, 1, 1);


            button_0.clicked.connect (() => { append_to_formula ("0"); });
            button_1.clicked.connect (() => { append_to_formula ("1"); });
            button_2.clicked.connect (() => { append_to_formula ("2"); });
            button_3.clicked.connect (() => { append_to_formula ("3"); });
            button_4.clicked.connect (() => { append_to_formula ("4"); });
            button_5.clicked.connect (() => { append_to_formula ("5"); });
            button_6.clicked.connect (() => { append_to_formula ("6"); });
            button_7.clicked.connect (() => { append_to_formula ("7"); });
            button_8.clicked.connect (() => { append_to_formula ("8"); });
		    button_9.clicked.connect (() => { append_to_formula ("9"); });

		    button_plus.clicked.connect (() => { append_to_formula ("+"); });
		    button_minus.clicked.connect (() => { append_to_formula ("-"); });
		    button_point.clicked.connect (() => { append_to_formula ("."); });
		    button_multi.clicked.connect (() => { append_to_formula ("*"); });
		    button_div.clicked.connect (() => { append_to_formula ("/"); });
		    button_percent.clicked.connect (() => { append_to_formula ("%"); });
		    button_clear.clicked.connect (() => { set_formula (""); });

		    button_history.clicked.connect (() => {
                if (menu.visible)
                    menu.popdown ();
                else {
                    menu = new Gtk.Menu ();
                    for (int i = 0; i < history.length; i++) {
                        string entry_formula = history.index (i).form;
                        string entry_result = history.index (i).resu;
                        string entry = parser.prettify_formula (entry_formula) + " = " + entry_result;
                        if (entry_result != "ERR") {
		                    var add_pl_menuitem = new Gtk.MenuItem.with_label (entry);
		                    add_pl_menuitem.activate.connect ( () => {
		                        paste_token (entry_result);
                            });
		                    menu.append (add_pl_menuitem);
		                }
		            }
		            menu.popup (null, null, null, Gdk.BUTTON_PRIMARY, Gtk.get_current_event_time ());
		            menu.show_all ();
                }
            });

		    button_equal.clicked.connect (() => { solve_equation (); });
        }

        private void set_formula (string new_formula, bool new_input_new_formula = false) {
            formula = new_formula;
            equation_changed ();
            input_new_formula = new_input_new_formula;
        }

        private void append_to_formula (string appendix) {
            if (input_new_formula) {
                if (!parser.is_operator (appendix))
                    formula = "";
                input_new_formula = false;
            }
            formula += appendix;
            equation_changed ();
        }


        private void solve_equation () {
            string result = parser.calculate_formula (formula);
            var entry = history_entry(){form = formula, resu = result};

            if (history.length > 0) {
                if (history.index(0) != entry && formula != result)
                    history.prepend_val (entry);
            } else
                history.prepend_val (entry);

            if (result == "ERR")
                show_error();
            else {
                // Add to History
                button_history.sensitive = true;
                set_formula (result, true);
            }
        }

        private void equation_changed () {
            set_formula_has_error (false);
            textview.buffer.text = parser.prettify_formula(formula);
        }

        private void show_error (string error_message = "") {
            set_formula_has_error (true);
        }


        private bool undo () {
            if (history.length > 0) {
                int current_state = -1;
                for (int i = 0; i < history.length; i++) {
                    string entry_formula = history.index(i).form;

                    if (formula == entry_formula) {
                        current_state = i;
                        break;
                    }
                }

                int next_state = current_state + 1;
                if (next_state >= history.length)
                    return false;

                string old_formula = history.index(next_state).form;
                set_formula (old_formula);

                return true;
            }
            return false;
        }

        private bool paste_token (string token) {
            string input = token.replace(" ", "");
            if (parser.is_formula_appendable(formula)) {
                append_to_formula (input);
            } else {
                set_formula (input, true);
            }
            return true;
        }

        private void set_formula_has_error (bool input) {
            formula_has_error = input;
            if (input) {
                var mybgcolor=Gdk.RGBA();
                mybgcolor.parse("rgba(180,20,20,1)");
                textview.override_color(Gtk.StateFlags.NORMAL, mybgcolor);
            } else
                textview.override_color(Gtk.StateFlags.NORMAL, null);
        }
    }
}