
/***
  Copyright (C) 2014 Elementary Calculator Lars Berscheid
  This program is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License version 3, as published
  by the Free Software Foundation.
  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranties of
  MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
  PURPOSE. See the GNU General Public License for more details.
  You should have received a copy of the GNU General Public License along
  with this program. If not, write to the
  Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.
***/


namespace Calculator {

    public errordomain ParserError {
        MISSING_PARE,
        MISSING_ARG,
        ZERO_DIVISION,
        UNKNOWN_TOKEN,
        MISSING_OPERATOR,
        EMPTY_FORMULA
    }

    public class Parser : GLib.Object {

        // Basic Structures for Parsing
        // prec...precedence, asso...associative={LEFT, RIGHT}
        private delegate double Evaluation (double a, double b = 0);
        private struct operator { string symbol; int prec; string asso; Evaluation eval; }
        private struct function { string symbol; string asso; Evaluation eval; }
        private struct constant { string symbol; double val; }

        // List of all Mathematical functions, any suggestions?
        private operator[] operators = { operator(){ symbol="+", prec=1, asso="LEFT", eval=(a,b) => { return a + b; }},
                                     operator(){ symbol="-", prec=1, asso="LEFT", eval=(a,b) => { return a - b; }},
                                     operator(){ symbol="*", prec=2, asso="LEFT", eval=(a,b) => { return a * b; }},
                                     operator(){ symbol="/", prec=2, asso="LEFT", eval=(a,b) => { return a / b; }},
                                     operator(){ symbol="mod", prec=2, asso="LEFT", eval=(a,b) => { return a % b; }},
                                     operator(){ symbol="**", prec=3, asso="RIGHT", eval=(a,b) => { return Math.pow(a,b); }},
                                     operator(){ symbol="e", prec=4, asso="RIGHT", eval=(a,b) => { return a*Math.pow(10,b); }},
                                     operator(){ symbol="%", prec=5, asso="LEFT", eval=(a,b) => { return a / 100; }} };

        // List of all Mathematical functions, any suggestions?
        private function[] functions = { function(){ symbol="sin", asso="RIGHT", eval=(a) => { return Math.sin(a); }},
                                     function(){ symbol="cos", asso="RIGHT", eval=(a) => { return Math.cos(a); }},
                                     function(){ symbol="tan", asso="RIGHT", eval=(a) => { return Math.tan(a); }},
                                     function(){ symbol="cot", asso="RIGHT", eval=(a) => { return 1 / Math.tan(a); }},
                                     function(){ symbol="sinh", asso="RIGHT", eval=(a) => { return Math.sinh(a); }},
                                     function(){ symbol="cosh", asso="RIGHT", eval=(a) => { return Math.cosh(a); }},
                                     function(){ symbol="exp", asso="RIGHT", eval=(a) => { return Math.exp(a); }},
                                     function(){ symbol="log", asso="RIGHT", eval=(a) => { return Math.log(a); }},
                                     function(){ symbol="sqrt", asso="RIGHT", eval=(a) => { return Math.sqrt(a); }} };

        // List of all Mathematical constants, e is difficult due to the exponential presentation
        private constant[] constants = { constant () { symbol = "pi", val = Math.PI }};

        private string decimal_symbol;

        public Parser (string new_decimal_symbol) {
            decimal_symbol = new_decimal_symbol;
        }

        // Increase readability of the formula
        public string prettify_formula (string formula) {
            string pretty_formula = formula;

            if (formula.length == 0)
                pretty_formula = "0";
            else {
                try {
                    Array<string> tokenized_formula = tokenize_formula (formula);

                    pretty_formula = "";
                    for (int i = 0; i < tokenized_formula.length; i++) {
                        string token = tokenized_formula.index (i);
                        if (is_operator (token)) {
                            if (token == "**")
                                token = "^";
                            else if (is_exponential (token))
                                token = "e";
                            else if (token == "%")
                                token = "%";
                            else if (i > 0 && is_plus (token) && is_exponential(tokenized_formula.index (i - 1)))
                                token = "+";
                            else
                                token = " " + token.replace ("*", "×").replace("/", "÷") + " ";
                        } else if (is_constant (token)) {
                            if (token == "pi")
                                token = "π";
                        } else if (is_function (token)) {
                            if (token == "sqrt")
                                token = "√";
                        } else {
                            if (decimal_symbol == ",")
                                token = token.replace (".", ",");
                        }
                        pretty_formula += token;
                    }
                } catch {
                    pretty_formula = pretty_formula.replace ("+", " + ");
                    pretty_formula = pretty_formula.replace ("-", " - ");
                    pretty_formula = pretty_formula.replace ("**", "^");
                    pretty_formula = pretty_formula.replace ("*", " × ");
                    pretty_formula = pretty_formula.replace ("/", " ÷ ");
                    pretty_formula = pretty_formula.replace ("pi", "π");
                    if (decimal_symbol == ",")
                        pretty_formula = pretty_formula.replace (".", ",");
                }
            }
            return pretty_formula;
        }

        // Overall process of calculation, outputs string!
        // input -> tokenize formula -> shift to reverse polish notation (rpn) -> calculate rpn -> round result
        // all private functions can throw errors, they are handled in here
        public string calculate_formula (string formula) {

            if (formula.length == 0)
                return "0";

            try {
                Array<string> tokenized_formula = tokenize_formula (formula, true);
                print (@"Tokenized Formula:");
                for (int i = 0; i < tokenized_formula.length; i++)
		            print (" %s |", tokenized_formula.index (i));
	            print (@"| \n");

                Array<string> rpn = tokenized_formula_to_rpn (tokenized_formula);
                print (@"RPN:");
                for (int i = 0; i < rpn.length; i++)
		            print (" %s |", rpn.index (i));
	            print (@"| \n");

                double result = calculate_rpn (rpn, tokenized_formula);
                print (@"Prerounded Result: $result\n");

                string result_string = round_result (result);
                print (@"Result: $result_string\n");
                return result_string;
            } catch (ParserError e) {
                print (@"Error: $(e.message) \n");
                return "ERR";
            }
        }

        public bool is_formula_appendable (string formula) {
            if (formula.length == 0)
                return false;

            try {
                Array<string> tokenized_formula = tokenize_formula (formula, true);
                if (tokenized_formula.length == 0)
                    return false;
                string last_token = tokenized_formula.index (tokenized_formula.length - 1);
                if (!is_number_or_constant (last_token))
                    return true;
                return false;
            } catch (ParserError e) {
                return false;
            }
        }


        // Split String into Tokens, save in Array
        private Array<string> tokenize_formula (string expr, bool remove_exponential_plus = false) throws ParserError {
            Array<string> input = new Array<string> ();

            int a = 0, b = 0;
	        while (b <= expr.length) {
	            for (int b_help = a; b_help <= expr.length; b_help++) {
                    if (is_token (expr[a:b_help], false, false))
                        b = b_help;

                    // Exit Condition if no Token was found
                    if (b_help == expr.length && b == a)
                        b = expr.length;
	            }

	            input.append_val( expr[a:b] );
	            a = b;
	            b++;
	        }

	        if (input.length == 0)
	            throw new ParserError.EMPTY_FORMULA("There is nothing in here.");

            // Handle Plus and Minus as Positive & Negative Sign
	        for (int i = 0; i < input.length; i++) {
	            string token = input.index (i);
	            if (is_plus (token) && remove_exponential_plus) {
	                if (i == 0) {
	                    input.remove_index (0);
	                } else {
	                    if (is_exponential (input.index(i - 1)) && is_number (input.index(i - 2))) {
	                        input.remove_index (i);
	                    }
	                }
	            }
	            if (is_minus (token)) {
	                if (i == 0) {
	                    if (input.length >= 2 && is_number_or_constant (input.index (i + 1))) {
	                        string following_number = input.index (i + 1);
                            input.remove_index (0);
                            input.remove_index (0);
                            input.prepend_val ("-" + following_number);
	                    }
	                } else {
	                    if (!is_number_or_constant (input.index (i - 1)) && is_number_or_constant (input.index (i + 1)) && input.index (i - 1) != "%") {
	                        string following_number = input.index (i + 1);
                            input.remove_index (i);
                            input.remove_index (i);
                            input.insert_val (i, "-" + following_number);
	                    }
	                }
	            }
	        }

	        return input;
        }

        // Use Shunting-Yard Algorithm to bring tokenized formula to rpn
        // More Information: https://en.wikipedia.org/wiki/Shunting-yard_algorithm
        private Array<string> tokenized_formula_to_rpn (Array<string> tokenized_formula) throws ParserError {
            Array<string> input = tokenized_formula;
            Array<string> output = new Array<string> ();
	        Array<string> stack = new Array<string> ();

            while (input.length > 0) {
                string token = input.index (0);
                input.remove_index (0);

                if (is_number_or_constant (token))
                    output.append_val (token);

                else if (is_function (token))
                    stack.append_val (token);

                else if (is_operator (token)) {
                    operator token_operator = get_operator (token);

                    if (stack.length > 0) {
                        string stack_end = stack.index (stack.length - 1);
                        if (is_operator (stack_end)) {
                            operator stack_end_operator = get_operator (stack_end);

                            while (stack.length > 0 && is_operator (stack_end) && (
                                    (token_operator.asso == "LEFT" && token_operator.prec <= stack_end_operator.prec) ||
                                    (token_operator.prec < stack_end_operator.prec)
                                )) {

                                output.append_val (stack_end);
                                stack.remove_index (stack.length - 1);

                                if (stack.length > 0) {
                                    stack_end = stack.index (stack.length - 1);
                                    if (is_operator (stack_end))
                                        stack_end_operator = get_operator (stack_end);
                                }
                            }
                        }
                    }

                    stack.append_val (token);
                }

                else if (is_left_pare (token))
                    stack.append_val (token);

                else if (is_right_pare (token)) {
                    if (stack.length == 0)
                        throw new ParserError.MISSING_PARE("There are more right than left pare.");

                    string stack_end = stack.index (stack.length - 1);

                    while (is_left_pare (stack_end) == false) {
                        output.append_val (stack_end);
                        stack.remove_index (stack.length - 1);

                        if (stack.length == 0)
                            throw new ParserError.MISSING_PARE("There are more right than left pare.");

                        stack_end = stack.index (stack.length - 1);
                    }
                    stack.remove_index (stack.length - 1);

                    if (stack.length > 0) {
                        stack_end = stack.index (stack.length - 1);
                        bool is_stack_end_function = is_function (stack_end);
                        if (is_stack_end_function) {
                            output.append_val (stack_end);
                            stack.remove_index (stack.length - 1);
                        }
                    }
                }

                else
                    throw new ParserError.UNKNOWN_TOKEN("Unknown Token in Formula.");
	        }

            while (stack.length > 0) {
                string stack_end = stack.index (stack.length - 1);

                if (is_left_pare (stack_end))
                    throw new ParserError.MISSING_PARE("There are more left than right pare.");

                output.append_val (stack_end);
                stack.remove_index (stack.length - 1);
            }

            return output;
        }

        // Calculate the numeric Result
        private double calculate_rpn (Array<string> rpn, Array<string> tokenized_formula) throws ParserError {
            Array<string> input = rpn;
            Array<string> output = new Array<string> ();

            // Replace the Constants
            for (int i = 0; i < input.length; i++) {
		        if (is_constant (input.index (i))) {
		            foreach (constant c in constants) {
                        if (input.index (i) == c.symbol) {
                            input.insert_val (i, @"$(c.val)");
		                    input.remove_index (i + 1);
		                }
                    }
		        }
		    }

            while (input.length > 0) {
                string token = input.index (0);
                input.remove_index (0);

                if (is_number (token))
                    output.append_val (token);

                if (is_function (token)) {
                    function token_function = get_function (token);

                    string token_1 = output.index (output.length - 1);
                    double double_1 = double.parse (token_1);

                    double result = token_function.eval (double_1);
                    output.remove_index (output.length - 1);
                    output.append_val (@"$result");
                }

                if (is_operator (token)) {
                    operator token_operator = get_operator (token);

                    if (token == "%") {
                        bool percentage_reference = false;
                        if (output.length < 1)
                            throw new ParserError.MISSING_ARG("No numbers for math operation.");

                        string token_1 = output.index (output.length - 1);
                        double double_1 = double.parse (token_1);

                        if (!is_number (token_1))
                            throw new ParserError.MISSING_ARG("No numbers for math operation.");

                        if (input.length > 0) {
                            if (input.index (0) == "+" || input.index (0) == "-") {
                                print ("Hererer\n");
                                percentage_reference = true;

                                string token_2 = output.index (output.length - 2);

                                if (!is_number (token_1) || !is_number (token_2))
                                    throw new ParserError.MISSING_ARG("No numbers for math operation.");

                                double double_2 = double.parse (token_2);

                                double result = double_1 * token_operator.eval (double_2);
                                output.remove_index (output.length - 1);
                                output.append_val ((@"$result").replace ("e+", "e"));
                            }
                        }

                        if (!percentage_reference) {
                            double result = token_operator.eval (double_1);
                            output.remove_index (output.length - 1);
                            output.append_val ((@"$result").replace ("e+", "e"));
                        }

                    } else {

                        if (output.length < 2)
                            throw new ParserError.MISSING_ARG("No numbers for math operation.");

                        string token_1 = output.index (output.length - 1);
                        string token_2 = output.index (output.length - 2);

                        if (!is_number (token_1) || !is_number (token_2))
                            throw new ParserError.MISSING_ARG("No numbers for math operation.");

                        double double_1 = double.parse (token_1);
                        double double_2 = double.parse (token_2);

                        if (token == "/" && double_1 == 0)
                            throw new ParserError.ZERO_DIVISION("There's a zero division going on.");

                        double result = token_operator.eval (double_2, double_1);
                        output.remove_index (output.length - 1);
                        output.remove_index (output.length - 1);
                        output.append_val ((@"$result").replace ("e+", "e"));
                    }
                }
            }

            if (output.length > 1)
                throw new ParserError.MISSING_OPERATOR("Not enough operators.");

            return double.parse (output.index (0));
        }

        // Round Result and convert it to a readable string
        // by Default to 6 decimal numbers
        private string round_result (double result, int max_round_length = 6) throws ParserError {
            string result_string = @"$result";
            string[] result_string_e_split = result_string.split ("e");
            double pre_e_split = double.parse (result_string_e_split[0]);
            string[] result_string_split = result_string_e_split[0].split (".");

            if (0 < Math.fabs (result) && Math.fabs (result) < 1e-14) {
                result_string = "0.";
                return result_string;
            }

            if (Math.fabs (Math.round (pre_e_split) - pre_e_split ) < Math.pow(10, -2 * max_round_length)) {
                result_string = @"$(Math.round (pre_e_split))";
                if (result_string_e_split.length == 2)
                    result_string += "e" + result_string_e_split[1];
                return result_string;
            }

            if (result_string_split.length == 2) {
                string pre_decimal = result_string_split[0];
                string post_decimal = result_string_split[1];
                if (post_decimal.length >= max_round_length) {

                    int count_null = 0;
                    for (int i = 0; i<max_round_length; i++) { if (post_decimal[i:i+1] == "0") { count_null++; } else { break; } }

                    double post_decimal_double = double.parse (post_decimal[0:max_round_length-1] + "." + post_decimal[max_round_length:post_decimal.length]);
                    double post_decimal_double_rounded = Math.round (post_decimal_double);
                    string post_decimal_new = @"$post_decimal_double_rounded";
                    if (post_decimal_double_rounded > post_decimal_double && post_decimal_new[0:1] == "1") {
                        count_null--;
                    }

                    if (post_decimal_new.length > max_round_length)
                        post_decimal_new = post_decimal_new[0:max_round_length];
                    while (post_decimal_new[-1:post_decimal_new.length] == "0") { post_decimal_new = post_decimal_new[0:-1]; }

                    result_string = pre_decimal + ".";
                    for (int i = 0; i<count_null; i++) { result_string += "0"; }
                    result_string += post_decimal_new;

                    if (result_string_e_split.length == 2)
                        result_string += "e" + result_string_e_split[1];
                }
            }

            return result_string;
        }


        // Helper Functions
        private operator get_operator (string operator_symbol) {
            foreach (operator o in operators) {
                if (operator_symbol == o.symbol)
                    return o;
            }
            return operator () { symbol="", prec=0, asso="LEFT", eval=(a,b) => { return a; }};
        }

        private function get_function (string function_symbol) {
            foreach (function f in functions) {
                if (function_symbol == f.symbol)
                    return f;
            }
            return function () { symbol="", eval = (a) => { return a; } };
        }


        public bool is_token (string input, bool negative_sign = true, bool exponential = true) {
            if (is_number(input, negative_sign, exponential) || is_constant (input, negative_sign) || is_operator(input) || is_function(input) || is_left_pare(input) || is_right_pare(input) )
                return true;
            return false;
        }

        public bool is_number (string input, bool negative_sign = true, bool exponential = true) {
            string match_string = input;
            if (input.length == 0)
                return false;
            if (is_exponential (input))
                return false;
            if (is_minus (input))
                return false;
            if (input[0:1] == "-" && input.length > 1 && negative_sign)
                match_string = input[1:input.length];
            if (exponential) {
                if (/^\d*\.?\d*e?-?\d*$/.match (match_string))
                    return true;
            } else {
                if (/^\d*\.?\d*$/.match (match_string))
                    return true;
            }

            return false;
        }

        public bool is_constant (string input, bool negative_sign = true) {
            string match_string = input;
            if (input.length == 0)
                return false;
            if (input[0:1] == "-" && input.length > 1 && negative_sign)
                match_string = input[1:input.length];
            foreach (constant c in constants) {
                if (match_string == c.symbol)
                    return true;
            }
            return false;
        }

        public bool is_number_or_constant (string input, bool negative_sign = true, bool exponential = true) {
            string match_string = input;
            if (input.length == 0)
                return false;
            if (input[0:1] == "-" && input.length > 1 && negative_sign)
                match_string = input[1:input.length];
            if (is_number (match_string) || is_constant (match_string))
                return true;
            return false;
        }

        public bool is_operator (string input) {
            foreach (operator o in operators) {
                if (input == o.symbol)
                    return true;
            }
            return false;
        }

        public bool is_function (string input) {
            foreach (function f in functions) {
                if (input == f.symbol)
                    return true;
            }
            return false;
        }

        public bool is_left_pare (string input) {
            if (input == "(")
                return true;
            return false;
        }

        public bool is_right_pare (string input) {
            if (input == ")")
                return true;
            return false;
        }

        public bool is_minus (string input) {
            if (input == "-")
                return true;
            return false;
        }

        public bool is_plus (string input) {
            if (input == "+")
                return true;
            return false;
        }

        public bool is_exponential (string input) {
            if (input == "e")
                return true;
            return false;
        }
    }
}